﻿using System;
using System.Runtime.InteropServices;
using System.Threading;

namespace Juniper.Ssi.WM
{
    /// <summary>
    ///    Provides a barcode reader for a motorolla ssi capable barcode reader device.
    /// </summary>
    public class BarcodeScanner : IDisposable
    {
        private const int MAX_WAIT_TIME = 10000;
        private const int SLEEP_WAIT_INTERVAL = 200;

        private readonly int _baudRate;
        private readonly int _portNumber;
        private readonly SsiInterceptor _ssiInterceptor;

        private byte[] _capabilitiesBuffer;
        private GCHandle _capabilitiesBufferHandle;
        private bool _connected;
        private bool _disposed;
        private volatile bool _waiting;
        private volatile bool _gotAnswer;
        private Thread _waitThread;

        private Action<bool> _failedCallback;
        private Action<BarcodeScannerCapabilities> _capabilitiesCallback;

        /// <summary>
        ///    Initializes a new instance of the <see cref = "BarcodeScanner" /> class.
        /// </summary>
        /// <param name = "portNumber">The port number.</param>
        /// <param name = "baudRate">The baud rate.</param>
        public BarcodeScanner(int portNumber, int baudRate)
        {
            _ssiInterceptor = new SsiInterceptor();
            _ssiInterceptor.CapabilitiesReturned += OnCapabilitiesReturned;
            //_ssiInterceptor.ErrorOccurred += OnErrorOccurred;

            _baudRate = baudRate;
            _portNumber = portNumber;
        }

        #region IDisposable Members

        /// <summary>
        ///    Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        #endregion

        public event EventHandler<BarcodeScannerDataReceivedEventArgs<BarcodeScannerCapabilities>> CapabilitiesReceived;

        /// <summary>
        ///    Opens this instance.
        /// </summary>
        public void Open()
        {
            if (_connected)
            {
                return;
            }

            if (_ssiInterceptor == null)
            {
                throw new InvalidOperationException("The scanner message intercepter is null.");
            }

            SsiInteropHelper.ThrowOnError(NativeMethods.SSIConnect(_ssiInterceptor.Hwnd, _baudRate, _portNumber));
            _connected = true;
        }

        /// <summary>
        ///    Closes this instance.
        /// </summary>
        public void Close()
        {
            if (!_connected)
            {
                return;
            }

            if (_ssiInterceptor == null)
            {
                throw new InvalidOperationException("The scanner message intercepter is null.");
            }

            SsiInteropHelper.ThrowOnError(NativeMethods.SSIDisconnect(_ssiInterceptor.Hwnd, _portNumber));
            _connected = false;
        }

        /// <summary>
        /// Called when [capabilities returned].
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        private void OnCapabilitiesReturned(object sender, EventArgs e)
        {
            _waiting = false;

            if (_capabilitiesBufferHandle.IsAllocated)
            {
                _capabilitiesBufferHandle.Free();
            }

            if (_capabilitiesCallback != null)
            {
                BarcodeScannerCapabilities capabilities = new BarcodeScannerCapabilities(_capabilitiesBuffer);
                _capabilitiesCallback(capabilities);
            }

            _capabilitiesCallback = null;
            //if (CapabilitiesReceived != null)
            //{
            //    BarcodeScannerCapabilities capabilities = new BarcodeScannerCapabilities(_capabilitiesBuffer);
            //    CapabilitiesReceived(this, new BarcodeScannerDataReceivedEventArgs<BarcodeScannerCapabilities>(capabilities));
            //}
        }

        /// <summary>
        /// Gets the scanner capabilities.
        /// </summary>
        public void GetScannerCapabilities(Action<bool> failureAction, Action<BarcodeScannerCapabilities> completedAction)
        {
            if (failureAction == null)
            {
                throw new ArgumentNullException("failureAction");
            }

            if (completedAction == null)
            {
                throw new ArgumentNullException("completedAction");
            }

            if (_waiting)
            {
                throw new InvalidOperationException("The barcode scanner is processing another command.");
            }

            _failedCallback = failureAction;
            _capabilitiesCallback = completedAction;

            StartWaiting();

            _capabilitiesBuffer = new byte[255];
            _capabilitiesBufferHandle = GCHandle.Alloc(_capabilitiesBuffer, GCHandleType.Pinned);

            try
            {
                SsiInteropHelper.ThrowOnError(NativeMethods.SetCapabilitiesBuffer(_portNumber, _capabilitiesBuffer, 255));
                SsiInteropHelper.ThrowOnError(NativeMethods.RequestScannerCapabilities(_portNumber));
            }
            catch (SsiException)
            {
                if (_capabilitiesBufferHandle.IsAllocated)
                {
                    _capabilitiesBufferHandle.Free();
                }

                _capabilitiesBuffer = null;
                _failedCallback = null;
                _capabilitiesCallback = null;

                throw;
            }
        }

        /// <summary>
        ///    Aims the on.
        /// </summary>
        public void AimOn()
        {
            SsiInteropHelper.ThrowOnError(NativeMethods.AimOn(_portNumber));
        }

        /// <summary>
        ///    Aims the off.
        /// </summary>
        public void AimOff()
        {
            SsiInteropHelper.ThrowOnError(NativeMethods.AimOff(_portNumber));
        }

        /// <summary>
        /// Beeps the specified beep code.
        /// </summary>
        /// <param name="beepCode">The beep code.</param>
        public void Beep(BeepCode beepCode)
        {
            SsiInteropHelper.ThrowOnError(NativeMethods.SoundBeeper(_portNumber, beepCode));
        }

        /// <summary>
        /// Gets the DLL version.
        /// </summary>
        /// <returns></returns>
        public Version GetDllVersion()
        {
            int version = NativeMethods.ReturnDLLVersion();

            int major = (version & 0x0000ff00) >> 8;
            int minor = version & 0x000000ff;

            return new Version(major, minor);
        }

        /// <summary>
        /// Starts the waiting.
        /// </summary>
        private void StartWaiting()
        {
            _waiting = true;
            _waitThread = new Thread(WaitProc);
            _waitThread.Name = "Barcode Scanner Response Wait Thread";
            _waitThread.Start();
        }

        /// <summary>
        /// Waits the proc.
        /// </summary>
        private void WaitProc()
        {
            int waitTime = 0;

            while (_waiting)
            {
                if (waitTime >= MAX_WAIT_TIME)
                {
                    _gotAnswer = false;
                    return;
                }

                Thread.Sleep(SLEEP_WAIT_INTERVAL);
                waitTime += SLEEP_WAIT_INTERVAL;
            }
        }

        /// <summary>
        ///    Releases unmanaged and - optionally - managed resources
        /// </summary>
        /// <param name = "disposing"><c>true</c> to release both managed and unmanaged resources; <c>false</c> to release only unmanaged resources.</param>
        protected virtual void Dispose(bool disposing)
        {
            if (disposing && !_disposed)
            {
                Close();

                _disposed = true;
            }
        }
    }
}