﻿using System;
using System.Collections.Generic;

namespace Juniper.Ssi.WM
{
   [Flags]
   public enum BarcodeScannerBaudRate : ushort
   {
      Baud300 = 0x0001,
      Baud600 = 0x0002,
      Baud1200 = 0x0004,
      Baud2400 = 0x0008,
      Baud4800 = 0x0010,
      Baud9600 = 0x0020,
      Baud19200 = 0x0040,
      Baud28800 = 0x0080,
      Baud38400 = 0x0100,
      Baud57600 = 0x0200,
      Baud115200 = 0x0400,
      Reserved1 = 0x0800,
      Reserved2 = 0x1000,
      Reserved3 = 0x2000,
      Reserved4 = 0x4000,
      Reserved5 = 0x8000
   }

   [Flags]
   public enum BarcodeScannerSerialParams : byte
   {
      OddParity = 0x0001,
      EvenParity = 0x0002,
      NoneParity = 0x0004,
      CheckParity = 0x0008,
      DoNotCheckParity = 0x0010,
      OneStopBit = 0x0020,
      TwoStopBits = 0x0040
   }

   [Flags]
   public enum BarcodeScannerMultiPacketOptions : byte
   {
      Option1 = 0x0001,
      Option2 = 0x0002,
      Option3 = 0x0004
   }

    public class BarcodeScannerCapabilities
    {
        private BarcodeScannerBaudRate _SupportedScannerBaudRates;
        private BarcodeScannerSerialParams _MiscScannerSerialParameters;
        private BarcodeScannerMultiPacketOptions _ScannerMultiPacketOptions;
        private IList<byte> _Commands;

        public BarcodeScannerCapabilities(byte[] capabilitiesBuffer)
        {
            if (capabilitiesBuffer == null)
            {
                throw new ArgumentNullException("capabilitiesBuffer");
            }

            Commands = new List<byte>();
            ParseBuffer(capabilitiesBuffer);
        }

        /// <summary>
        /// Parses the buffer.
        /// </summary>
        /// <param name="buffer">The buffer.</param>
        private void ParseBuffer(byte[] buffer)
        {
            ParseSupportedBaudRates(buffer);
            ParseMiscSerialParameters(buffer);
            ParseMultiPacketOptions(buffer);
            ParseCommands(buffer);
        }

        /// <summary>
        /// Parses the supported baud rates.
        /// </summary>
        /// <param name="buffer">The buffer.</param>
        private void ParseSupportedBaudRates(byte[] buffer)
        {
            ushort field = BitConverter.ToUInt16(buffer, 0);
            SupportedScannerBaudRates = (BarcodeScannerBaudRate)Enum.ToObject(typeof(BarcodeScannerBaudRate), field);
        }

        /// <summary>
        /// Parses the misc serial parameters.
        /// </summary>
        /// <param name="buffer">The buffer.</param>
        private void ParseMiscSerialParameters(byte[] buffer)
        {
            byte field = buffer[2];
            MiscScannerSerialParameters = (BarcodeScannerSerialParams)Enum.ToObject(typeof(BarcodeScannerSerialParams), field);
        }

        /// <summary>
        /// Parses the multi packet options.
        /// </summary>
        /// <param name="buffer">The buffer.</param>
        private void ParseMultiPacketOptions(byte[] buffer)
        {
            byte field = buffer[3];
            ScannerMultiPacketOptions = (BarcodeScannerMultiPacketOptions)Enum.ToObject(typeof(BarcodeScannerMultiPacketOptions), field);
        }

        /// <summary>
        /// Parses the commands.
        /// </summary>
        /// <param name="buffer">The buffer.</param>
        private void ParseCommands(byte[] buffer)
        {
            for (int commandIndex = 4; commandIndex < buffer.Length; commandIndex++)
            {
                Commands.Add(buffer[commandIndex]);
            }
        }

        public BarcodeScannerBaudRate SupportedScannerBaudRates { 
            get
            {
                return _SupportedScannerBaudRates;
            }
            private set
            {
                _SupportedScannerBaudRates = value;
            }
        }

        public BarcodeScannerSerialParams MiscScannerSerialParameters
        {
            get
            {
                return _MiscScannerSerialParameters;
            }
            private set
            {
                _MiscScannerSerialParameters = value;
            }
        }

        public BarcodeScannerMultiPacketOptions ScannerMultiPacketOptions
        {
            get
            {
                return _ScannerMultiPacketOptions;
            }
            private set
            {
                _ScannerMultiPacketOptions = value;
            }
        }

        public IList<byte> Commands 
        {
            get
            {
                return _Commands;
            }
            private set
            {
                _Commands = value;
            }
        }
    }
}